@@ -1,28 +1,33 @@
+use std::ffi::OsStr;
 use std::fs::{self, File};
 use std::io::{self, Write};
 use std::path::{Path, PathBuf};
 
-use chrono::{Datelike, Timelike};
+use chrono::{DateTime, Datelike, Timelike, Utc};
 use codespan_reporting::diagnostic::{Diagnostic, Label};
 use codespan_reporting::term;
-use ecow::{eco_format, EcoString};
+use ecow::eco_format;
 use parking_lot::RwLock;
+use pathdiff::diff_paths;
 use rayon::iter::{IntoParallelRefIterator, ParallelIterator};
 use typst::diag::{
     bail, At, Severity, SourceDiagnostic, SourceResult, StrResult, Warned,
 };
 use typst::foundations::{Datetime, Smart};
-use typst::layout::{Frame, Page, PageRanges};
-use typst::model::Document;
+use typst::html::HtmlDocument;
+use typst::layout::{Frame, Page, PageRanges, PagedDocument};
 use typst::syntax::{FileId, Source, Span};
 use typst::WorldExt;
-use typst_pdf::{PdfOptions, PdfStandards};
+use typst_pdf::{PdfOptions, PdfStandards, Timestamp};
 
 use crate::args::{
-    CompileCommand, DiagnosticFormat, Input, Output, OutputFormat, PageRangeArgument,
-    PdfStandard,
+    CompileArgs, CompileCommand, DiagnosticFormat, Input, Output, OutputFormat,
+    PdfStandard, WatchCommand,
 };
+#[cfg(feature = "http-server")]
+use crate::server::HtmlServer;
 use crate::timings::Timer;
+
 use crate::watch::Status;
 use crate::world::SystemWorld;
 use crate::{set_failed, terminal};
@@ -30,34 +35,73 @@ use crate::{set_failed, terminal};
 type CodespanResult<T> = Result<T, CodespanError>;
 type CodespanError = codespan_reporting::files::Error;
 
-impl CompileCommand {
-    /// The output path.
-    pub fn output(&self) -> Output {
-        self.output.clone().unwrap_or_else(|| {
-            let Input::Path(path) = &self.common.input else {
-                panic!("output must be specified when input is from stdin, as guarded by the CLI");
-            };
-            Output::Path(path.with_extension(
-                match self.output_format().unwrap_or(OutputFormat::Pdf) {
-                    OutputFormat::Pdf => "pdf",
-                    OutputFormat::Png => "png",
-                    OutputFormat::Svg => "svg",
-                },
-            ))
-        })
+/// Execute a compilation command.
+pub fn compile(timer: &mut Timer, command: &CompileCommand) -> StrResult<()> {
+    let mut config = CompileConfig::new(command)?;
+    let mut world =
+        SystemWorld::new(&command.args.input, &command.args.world, &command.args.process)
+            .map_err(|err| eco_format!("{err}"))?;
+    timer.record(&mut world, |world| compile_once(world, &mut config))?
+}
+
+/// A preprocessed `CompileCommand`.
+pub struct CompileConfig {
+    /// Whether we are watching.
+    pub watching: bool,
+    /// Path to input Typst file or stdin.
+    pub input: Input,
+    /// Path to output file (PDF, PNG, SVG, or HTML).
+    pub output: Output,
+    /// The format of the output file.
+    pub output_format: OutputFormat,
+    /// Which pages to export.
+    pub pages: Option<PageRanges>,
+    /// The document's creation date formatted as a UNIX timestamp, with UTC suffix.
+    pub creation_timestamp: Option<DateTime<Utc>>,
+    /// The format to emit diagnostics in.
+    pub diagnostic_format: DiagnosticFormat,
+    /// Opens the output file with the default viewer or a specific program after
+    /// compilation.
+    pub open: Option<Option<String>>,
+    /// One (or multiple comma-separated) PDF standards that Typst will enforce
+    /// conformance with.
+    pub pdf_standards: PdfStandards,
+    /// A path to write a Makefile rule describing the current compilation.
+    pub make_deps: Option<PathBuf>,
+    /// The PPI (pixels per inch) to use for PNG export.
+    pub ppi: f32,
+    /// The export cache for images, used for caching output files in `typst
+    /// watch` sessions with images.
+    pub export_cache: ExportCache,
+    /// Server for `typst watch` to HTML.
+    #[cfg(feature = "http-server")]
+    pub server: Option<HtmlServer>,
+}
+
+impl CompileConfig {
+    /// Preprocess a `CompileCommand`, producing a compilation config.
+    pub fn new(command: &CompileCommand) -> StrResult<Self> {
+        Self::new_impl(&command.args, None)
     }
 
-    /// The format to use for generated output, either specified by the user or inferred from the extension.
-    ///
-    /// Will return `Err` if the format was not specified and could not be inferred.
-    pub fn output_format(&self) -> StrResult<OutputFormat> {
-        Ok(if let Some(specified) = self.format {
+    /// Preprocess a `WatchCommand`, producing a compilation config.
+    pub fn watching(command: &WatchCommand) -> StrResult<Self> {
+        Self::new_impl(&command.args, Some(command))
+    }
+
+    /// The shared implementation of [`CompileConfig::new`] and
+    /// [`CompileConfig::watching`].
+    fn new_impl(args: &CompileArgs, watch: Option<&WatchCommand>) -> StrResult<Self> {
+        let input = args.input.clone();
+
+        let output_format = if let Some(specified) = args.format {
             specified
-        } else if let Some(Output::Path(output)) = &self.output {
+        } else if let Some(Output::Path(output)) = &args.output {
             match output.extension() {
                 Some(ext) if ext.eq_ignore_ascii_case("pdf") => OutputFormat::Pdf,
                 Some(ext) if ext.eq_ignore_ascii_case("png") => OutputFormat::Png,
                 Some(ext) if ext.eq_ignore_ascii_case("svg") => OutputFormat::Svg,
+                Some(ext) if ext.eq_ignore_ascii_case("html") => OutputFormat::Html,
                 _ => bail!(
                     "could not infer output format for path {}.\n\
                      consider providing the format manually with `--format/-f`",
@@ -66,43 +110,66 @@ impl CompileCommand {
             }
         } else {
             OutputFormat::Pdf
+        };
+
+        let output = args.output.clone().unwrap_or_else(|| {
+            let Input::Path(path) = &input else {
+                panic!("output must be specified when input is from stdin, as guarded by the CLI");
+            };
+            Output::Path(path.with_extension(
+                match output_format {
+                    OutputFormat::Pdf => "pdf",
+                    OutputFormat::Png => "png",
+                    OutputFormat::Svg => "svg",
+                    OutputFormat::Html => "html",
+                },
+            ))
+        });
+
+        let pages = args.pages.as_ref().map(|export_ranges| {
+            PageRanges::new(export_ranges.iter().map(|r| r.0.clone()).collect())
+        });
+
+        let pdf_standards = {
+            let list = args
+                .pdf_standard
+                .iter()
+                .map(|standard| match standard {
+                    PdfStandard::V_1_7 => typst_pdf::PdfStandard::V_1_7,
+                    PdfStandard::A_2b => typst_pdf::PdfStandard::A_2b,
+                    PdfStandard::A_3b => typst_pdf::PdfStandard::A_3b,
+                })
+                .collect::<Vec<_>>();
+            PdfStandards::new(&list)?
+        };
+
+        #[cfg(feature = "http-server")]
+        let server = match watch {
+            Some(command)
+                if output_format == OutputFormat::Html && !command.server.no_serve =>
+            {
+                Some(HtmlServer::new(&input, &command.server)?)
+            }
+            _ => None,
+        };
+
+        Ok(Self {
+            watching: watch.is_some(),
+            input,
+            output,
+            output_format,
+            pages,
+            pdf_standards,
+            creation_timestamp: args.world.creation_timestamp,
+            make_deps: args.make_deps.clone(),
+            ppi: args.ppi,
+            diagnostic_format: args.process.diagnostic_format,
+            open: args.open.clone(),
+            export_cache: ExportCache::new(),
+            #[cfg(feature = "http-server")]
+            server,
         })
     }
-
-    /// The ranges of the pages to be exported as specified by the user.
-    ///
-    /// This returns `None` if all pages should be exported.
-    pub fn exported_page_ranges(&self) -> Option<PageRanges> {
-        self.pages.as_ref().map(|export_ranges| {
-            PageRanges::new(
-                export_ranges.iter().map(PageRangeArgument::to_range).collect(),
-            )
-        })
-    }
-
-    /// The PDF standards to try to conform with.
-    pub fn pdf_standards(&self) -> StrResult<PdfStandards> {
-        let list = self
-            .pdf_standard
-            .iter()
-            .map(|standard| match standard {
-                PdfStandard::V_1_7 => typst_pdf::PdfStandard::V_1_7,
-                PdfStandard::A_2b => typst_pdf::PdfStandard::A_2b,
-            })
-            .collect::<Vec<_>>();
-        PdfStandards::new(&list)
-    }
-}
-
-/// Execute a compilation command.
-pub fn compile(mut timer: Timer, mut command: CompileCommand) -> StrResult<()> {
-    // Only meant for input validation
-    _ = command.output_format()?;
-
-    let mut world =
-        SystemWorld::new(&command.common).map_err(|err| eco_format!("{err}"))?;
-    timer.record(&mut world, |world| compile_once(world, &mut command, false))??;
-    Ok(())
 }
 
 /// Compile a single time.
@@ -111,96 +178,132 @@ pub fn compile(mut timer: Timer, mut command: CompileCommand) -> StrResult<()> {
 #[typst_macros::time(name = "compile once")]
 pub fn compile_once(
     world: &mut SystemWorld,
-    command: &mut CompileCommand,
-    watching: bool,
+    config: &mut CompileConfig,
 ) -> StrResult<()> {
     let start = std::time::Instant::now();
-    if watching {
-        Status::Compiling.print(command).unwrap();
+    if config.watching {
+        Status::Compiling.print(config).unwrap();
     }
 
-    let Warned { output, warnings } = typst::compile(world);
-    let result = output.and_then(|document| export(world, &document, command, watching));
+    let Warned { output, warnings } = compile_and_export(world, config);
 
-    match result {
+    match output {
         // Export the PDF / PNG.
-        Ok(()) => {
+        Ok(outputs) => {
             let duration = start.elapsed();
 
-            if watching {
+            if config.watching {
                 if warnings.is_empty() {
-                    Status::Success(duration).print(command).unwrap();
+                    Status::Success(duration).print(config).unwrap();
                 } else {
-                    Status::PartialSuccess(duration).print(command).unwrap();
+                    Status::PartialSuccess(duration).print(config).unwrap();
                 }
             }
 
-            print_diagnostics(world, &[], &warnings, command.common.diagnostic_format)
+            print_diagnostics(world, &[], &warnings, config.diagnostic_format)
                 .map_err(|err| eco_format!("failed to print diagnostics ({err})"))?;
 
-            write_make_deps(world, command)?;
-
-            if let Some(open) = command.open.take() {
-                if let Output::Path(file) = command.output() {
-                    open_file(open.as_deref(), &file)?;
-                }
-            }
+            write_make_deps(world, config, outputs)?;
+            open_output(config)?;
         }
 
         // Print diagnostics.
         Err(errors) => {
             set_failed();
 
-            if watching {
-                Status::Error.print(command).unwrap();
+            if config.watching {
+                Status::Error.print(config).unwrap();
             }
 
-            print_diagnostics(
-                world,
-                &errors,
-                &warnings,
-                command.common.diagnostic_format,
-            )
-            .map_err(|err| eco_format!("failed to print diagnostics ({err})"))?;
+            print_diagnostics(world, &errors, &warnings, config.diagnostic_format)
+                .map_err(|err| eco_format!("failed to print diagnostics ({err})"))?;
         }
     }
 
     Ok(())
 }
 
-/// Export into the target format.
-fn export(
+/// Compile and then export the document.
+fn compile_and_export(
     world: &mut SystemWorld,
-    document: &Document,
-    command: &CompileCommand,
-    watching: bool,
-) -> SourceResult<()> {
-    match command.output_format().at(Span::detached())? {
+    config: &mut CompileConfig,
+) -> Warned<SourceResult<Vec<Output>>> {
+    match config.output_format {
+        OutputFormat::Html => {
+            let Warned { output, warnings } = typst::compile::<HtmlDocument>(world);
+            let result = output.and_then(|document| export_html(&document, config));
+            Warned {
+                output: result.map(|()| vec![config.output.clone()]),
+                warnings,
+            }
+        }
+        _ => {
+            let Warned { output, warnings } = typst::compile::<PagedDocument>(world);
+            let result = output.and_then(|document| export_paged(&document, config));
+            Warned { output: result, warnings }
+        }
+    }
+}
+
+/// Export to HTML.
+fn export_html(document: &HtmlDocument, config: &CompileConfig) -> SourceResult<()> {
+    let html = typst_html::html(document)?;
+    let result = config.output.write(html.as_bytes());
+
+    #[cfg(feature = "http-server")]
+    if let Some(server) = &config.server {
+        server.update(html);
+    }
+
+    result
+        .map_err(|err| eco_format!("failed to write HTML file ({err})"))
+        .at(Span::detached())
+}
+
+/// Export to a paged target format.
+fn export_paged(
+    document: &PagedDocument,
+    config: &CompileConfig,
+) -> SourceResult<Vec<Output>> {
+    match config.output_format {
+        OutputFormat::Pdf => {
+            export_pdf(document, config).map(|()| vec![config.output.clone()])
+        }
         OutputFormat::Png => {
-            export_image(world, document, command, watching, ImageExportFormat::Png)
-                .at(Span::detached())
+            export_image(document, config, ImageExportFormat::Png).at(Span::detached())
         }
         OutputFormat::Svg => {
-            export_image(world, document, command, watching, ImageExportFormat::Svg)
-                .at(Span::detached())
+            export_image(document, config, ImageExportFormat::Svg).at(Span::detached())
         }
-        OutputFormat::Pdf => export_pdf(document, command),
+        OutputFormat::Html => unreachable!(),
     }
 }
 
 /// Export to a PDF.
-fn export_pdf(document: &Document, command: &CompileCommand) -> SourceResult<()> {
+fn export_pdf(document: &PagedDocument, config: &CompileConfig) -> SourceResult<()> {
+    // If the timestamp is provided through the CLI, use UTC suffix,
+    // else, use the current local time and timezone.
+    let timestamp = match config.creation_timestamp {
+        Some(timestamp) => convert_datetime(timestamp).map(Timestamp::new_utc),
+        None => {
+            let local_datetime = chrono::Local::now();
+            convert_datetime(local_datetime).and_then(|datetime| {
+                Timestamp::new_local(
+                    datetime,
+                    local_datetime.offset().local_minus_utc() / 60,
+                )
+            })
+        }
+    };
     let options = PdfOptions {
         ident: Smart::Auto,
-        timestamp: convert_datetime(
-            command.common.creation_timestamp.unwrap_or_else(chrono::Utc::now),
-        ),
-        page_ranges: command.exported_page_ranges(),
-        standards: command.pdf_standards().at(Span::detached())?,
+        timestamp,
+        page_ranges: config.pages.clone(),
+        standards: config.pdf_standards.clone(),
     };
     let buffer = typst_pdf::pdf(document, &options)?;
-    command
-        .output()
+    config
+        .output
         .write(&buffer)
         .map_err(|err| eco_format!("failed to write PDF file ({err})"))
         .at(Span::detached())?;
@@ -208,7 +311,9 @@ fn export_pdf(document: &Document, command: &CompileCommand) -> SourceResult<()>
 }
 
 /// Convert [`chrono::DateTime`] to [`Datetime`]
-fn convert_datetime(date_time: chrono::DateTime<chrono::Utc>) -> Option<Datetime> {
+fn convert_datetime<Tz: chrono::TimeZone>(
+    date_time: chrono::DateTime<Tz>,
+) -> Option<Datetime> {
     Datetime::from_ymd_hms(
         date_time.year(),
         date_time.month().try_into().ok()?,
@@ -228,36 +333,31 @@ enum ImageExportFormat {
 
 /// Export to one or multiple images.
 fn export_image(
-    world: &mut SystemWorld,
-    document: &Document,
-    command: &CompileCommand,
-    watching: bool,
+    document: &PagedDocument,
+    config: &CompileConfig,
     fmt: ImageExportFormat,
-) -> StrResult<()> {
-    let output = command.output();
+) -> StrResult<Vec<Output>> {
     // Determine whether we have indexable templates in output
-    let can_handle_multiple = match output {
+    let can_handle_multiple = match config.output {
         Output::Stdout => false,
         Output::Path(ref output) => {
             output_template::has_indexable_template(output.to_str().unwrap_or_default())
         }
     };
 
-    let exported_page_ranges = command.exported_page_ranges();
-
     let exported_pages = document
         .pages
         .iter()
         .enumerate()
         .filter(|(i, _)| {
-            exported_page_ranges.as_ref().map_or(true, |exported_page_ranges| {
+            config.pages.as_ref().map_or(true, |exported_page_ranges| {
                 exported_page_ranges.includes_page_index(*i)
             })
         })
         .collect::<Vec<_>>();
 
     if !can_handle_multiple && exported_pages.len() > 1 {
-        let err = match output {
+        let err = match config.output {
             Output::Stdout => "to stdout",
             Output::Path(_) => {
                 "without a page number template ({p}, {0p}) in the output path"
@@ -266,15 +366,13 @@ fn export_image(
         bail!("cannot export multiple images {err}");
     }
 
-    let cache = world.export_cache();
-
     // The results are collected in a `Vec<()>` which does not allocate.
     exported_pages
         .par_iter()
         .map(|(i, page)| {
             // Use output with converted path.
-            let output = match output {
-                Output::Path(ref path) => {
+            let output = match &config.output {
+                Output::Path(path) => {
                     let storage;
                     let path = if can_handle_multiple {
                         storage = output_template::format(
@@ -290,8 +388,11 @@ fn export_image(
                     // If we are not watching, don't use the cache.
                     // If the frame is in the cache, skip it.
                     // If the file does not exist, always create it.
-                    if watching && cache.is_cached(*i, &page.frame) && path.exists() {
-                        return Ok(());
+                    if config.watching
+                        && config.export_cache.is_cached(*i, &page.frame)
+                        && path.exists()
+                    {
+                        return Ok(Output::Path(path.to_path_buf()));
                     }
 
                     Output::Path(path.to_owned())
@@ -299,12 +400,10 @@ fn export_image(
                 Output::Stdout => Output::Stdout,
             };
 
-            export_image_page(command, page, &output, fmt)?;
-            Ok(())
+            export_image_page(config, page, &output, fmt)?;
+            Ok(output)
         })
-        .collect::<Result<Vec<()>, EcoString>>()?;
-
-    Ok(())
+        .collect::<StrResult<Vec<Output>>>()
 }
 
 mod output_template {
@@ -338,14 +437,14 @@ mod output_template {
 
 /// Export single image.
 fn export_image_page(
-    command: &CompileCommand,
+    config: &CompileConfig,
     page: &Page,
     output: &Output,
     fmt: ImageExportFormat,
 ) -> StrResult<()> {
     match fmt {
         ImageExportFormat::Png => {
-            let pixmap = typst_render::render(page, command.ppi / 72.0);
+            let pixmap = typst_render::render(page, config.ppi / 72.0);
             let buf = pixmap
                 .encode_png()
                 .map_err(|err| eco_format!("failed to encode PNG file ({err})"))?;
@@ -409,14 +508,25 @@ impl ExportCache {
 /// Writes a Makefile rule describing the relationship between the output and
 /// its dependencies to the path specified by the --make-deps argument, if it
 /// was provided.
-fn write_make_deps(world: &mut SystemWorld, command: &CompileCommand) -> StrResult<()> {
-    let Some(ref make_deps_path) = command.make_deps else { return Ok(()) };
-    let Output::Path(output_path) = command.output() else {
-        bail!("failed to create make dependencies file because output was stdout")
-    };
-    let Ok(output_path) = output_path.into_os_string().into_string() else {
+fn write_make_deps(
+    world: &mut SystemWorld,
+    config: &CompileConfig,
+    outputs: Vec<Output>,
+) -> StrResult<()> {
+    let Some(ref make_deps_path) = config.make_deps else { return Ok(()) };
+    let Ok(output_paths) = outputs
+        .into_iter()
+        .filter_map(|o| match o {
+            Output::Path(path) => Some(path.into_os_string().into_string()),
+            Output::Stdout => None,
+        })
+        .collect::<Result<Vec<_>, _>>()
+    else {
         bail!("failed to create make dependencies file because output path was not valid unicode")
     };
+    if output_paths.is_empty() {
+        bail!("failed to create make dependencies file because output was stdout")
+    }
 
     // Based on `munge` in libcpp/mkdeps.cc from the GCC source code. This isn't
     // perfect as some special characters can't be escaped.
@@ -430,6 +540,10 @@ fn write_make_deps(world: &mut SystemWorld, command: &CompileCommand) -> StrResu
                     res.push('$');
                     slashes = 0;
                 }
+                ':' => {
+                    res.push('\\');
+                    slashes = 0;
+                }
                 ' ' | '\t' => {
                     // `munge`'s source contains a comment here that says: "A
                     // space or tab preceded by 2N+1 backslashes represents N
@@ -452,18 +566,29 @@ fn write_make_deps(world: &mut SystemWorld, command: &CompileCommand) -> StrResu
 
     fn write(
         make_deps_path: &Path,
-        output_path: String,
+        output_paths: Vec<String>,
         root: PathBuf,
         dependencies: impl Iterator<Item = PathBuf>,
     ) -> io::Result<()> {
         let mut file = File::create(make_deps_path)?;
+        let current_dir = std::env::current_dir()?;
+        let relative_root = diff_paths(&root, &current_dir).unwrap_or(root.clone());
 
-        file.write_all(munge(&output_path).as_bytes())?;
+        for (i, output_path) in output_paths.into_iter().enumerate() {
+            if i != 0 {
+                file.write_all(b" ")?;
+            }
+            file.write_all(munge(&output_path).as_bytes())?;
+        }
         file.write_all(b":")?;
         for dependency in dependencies {
-            let Some(dependency) =
-                dependency.strip_prefix(&root).unwrap_or(&dependency).to_str()
-            else {
+            let relative_dependency = match dependency.strip_prefix(&root) {
+                Ok(root_relative_dependency) => {
+                    relative_root.join(root_relative_dependency)
+                }
+                Err(_) => dependency,
+            };
+            let Some(relative_dependency) = relative_dependency.to_str() else {
                 // Silently skip paths that aren't valid unicode so we still
                 // produce a rule that will work for the other paths that can be
                 // processed.
@@ -471,34 +596,50 @@ fn write_make_deps(world: &mut SystemWorld, command: &CompileCommand) -> StrResu
             };
 
             file.write_all(b" ")?;
-            file.write_all(munge(dependency).as_bytes())?;
+            file.write_all(munge(relative_dependency).as_bytes())?;
         }
         file.write_all(b"\n")?;
 
         Ok(())
     }
 
-    write(make_deps_path, output_path, world.root().to_owned(), world.dependencies())
+    write(make_deps_path, output_paths, world.root().to_owned(), world.dependencies())
         .map_err(|err| {
             eco_format!("failed to create make dependencies file due to IO error ({err})")
         })
 }
 
-/// Opens the given file using:
-/// - The default file viewer if `open` is `None`.
-/// - The given viewer provided by `open` if it is `Some`.
-///
-/// If the file could not be opened, an error is returned.
-fn open_file(open: Option<&str>, path: &Path) -> StrResult<()> {
+/// Opens the output if desired.
+fn open_output(config: &mut CompileConfig) -> StrResult<()> {
+    let Some(viewer) = config.open.take() else { return Ok(()) };
+
+    #[cfg(feature = "http-server")]
+    if let Some(server) = &config.server {
+        let url = format!("http://{}", server.addr());
+        return open_path(OsStr::new(&url), viewer.as_deref());
+    }
+
+    // Can't open stdout.
+    let Output::Path(path) = &config.output else { return Ok(()) };
+
     // Some resource openers require the path to be canonicalized.
     let path = path
         .canonicalize()
         .map_err(|err| eco_format!("failed to canonicalize path ({err})"))?;
-    if let Some(app) = open {
-        open::with_detached(&path, app)
-            .map_err(|err| eco_format!("failed to open file with {} ({})", app, err))
+
+    open_path(path.as_os_str(), viewer.as_deref())
+}
+
+/// Opens the given file using:
+///
+/// - The default file viewer if `app` is `None`.
+/// - The given viewer provided by `app` if it is `Some`.
+fn open_path(path: &OsStr, viewer: Option<&str>) -> StrResult<()> {
+    if let Some(viewer) = viewer {
+        open::with_detached(path, viewer)
+            .map_err(|err| eco_format!("failed to open file with {} ({})", viewer, err))
     } else {
-        open::that_detached(&path).map_err(|err| {
+        open::that_detached(path).map_err(|err| {
             let openers = open::commands(path)
                 .iter()
                 .map(|command| command.get_program().to_string_lossy())
